//----------------------------------------------------------------------------------
// File:   CkTools.h
// Author: Samuel Gateau
// Email:  sdkfeedback@nvidia.com
// 
// Copyright (c) 2007 NVIDIA Corporation. All rights reserved.
//
// TO  THE MAXIMUM  EXTENT PERMITTED  BY APPLICABLE  LAW, THIS SOFTWARE  IS PROVIDED
// *AS IS*  AND NVIDIA AND  ITS SUPPLIERS DISCLAIM  ALL WARRANTIES,  EITHER  EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED  TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL  NVIDIA OR ITS SUPPLIERS
// BE  LIABLE  FOR  ANY  SPECIAL,  INCIDENTAL,  INDIRECT,  OR  CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION,  DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE  USE OF OR INABILITY  TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
//
//
//----------------------------------------------------------------------------------


// helper classes for demo

#pragma once

#pragma warning(disable:4995) // avoird warning for "...was marked as #pragma deprecated"
#pragma warning(disable:4239) // avoird warning for "...nonstandard extension used"

#include "DXUTgui.h"
#include "DXUTsettingsDlg.h"
#include "sdkmisc.h"

#include <string>
#include <vector>

HWND g_hWnd;

/** convert string to wstring
 */
std::wstring& stringtowstring(const std::string& pString, std::wstring& pWString)
{
    for (unsigned int charId = 0; charId < pString.size(); ++charId)
    {
        pWString += (WCHAR) pString[charId];
    }
    return pWString;
}

#define GET_MOUSECURSORX(pParam)  static_cast<int>(LOWORD(pParam))
#define GET_MOUSECURSORY(pParam)  static_cast<int>(HIWORD(pParam))
#define GET_BUTTONSTATE(pParam)  static_cast<int>(LOWORD(pParam))

/** CkCamera
 */
class CkCamera
{
public:
    /********************************* TYPES **********************************/

    /** Camera modes.
     */
    enum CkMode
    {
        eInspect = 0,
        eTerrain,
    };
    
    /** Projection struct.
     */
    struct CkProjection
    {
        float mAperture;
        float mAspectRatio;
        float mNear;
        float mFar;
    };

    /** Input State struct.
     */
    struct CkState
    {
        int mButtonState;
        int mMouseCursorX;
        int mMouseCursorY;
        int mMouseWheelDelta;
        unsigned int mKeyCode;

        CkState():
            mButtonState(0),
            mMouseCursorX(0),
            mMouseCursorY(0),
            mMouseWheelDelta(0),
            mKeyCode(0)
            {}
        CkState(int pButtonState, int pCursosX, int pCursorY, int pWheel = 0):
            mButtonState(pButtonState),
            mMouseCursorX(pCursosX),
            mMouseCursorY(pCursorY),
            mMouseWheelDelta(pWheel)
            {}
        CkState(int pButtonState, unsigned int pKeyCode):
            mButtonState(pButtonState),
            mKeyCode(pKeyCode)
            {}
    };
    
    /******************************* ATTRIBUTES *******************************/

    /******************************** METHODS *********************************/

    /** Constructor - Destructor.
     */
    CkCamera(CkMode pMode = eInspect) :
        mMode(pMode),
        mCamInspectPoint(0, 0, 0),
        mCamTranslate(0, 0, 0),
        mCamRotate(0, 0, 0),
        mCamLocalPos(0, 0, -10),
        mCamSpeed(0.005f, 0.005f, 0.1f)
    {
        setProjectionFrame(0.76f, 1.f, 0.01f, 100.f);
        setViewFrame(mCamInspectPoint, mCamRotate, mCamLocalPos.z, true);
        updateWorldFrame();
    }
    virtual ~CkCamera() {}

    /** Access the different frames of the Camera expressed in world coordinate.
     */
    const D3DXMATRIX& getProjectionFrame() const {return mProjectionFrame;}
    const D3DXMATRIX& getViewFrame() const {return mViewFrame;}
    const D3DXMATRIX& getWorldFrame() const {return mWorldFrame;}

    D3DXMATRIX& getWorldViewFrame(D3DXMATRIX& pFrame) const 
    {
        D3DXMatrixMultiply(&pFrame, &getWorldFrame(), &getViewFrame());
        return pFrame;
    }

    D3DXMATRIX& getWorldViewProjectionFrame(D3DXMATRIX& pFrame) const 
    {
        D3DXMatrixMultiply(&pFrame, &getWorldFrame(), &getViewFrame());
        D3DXMatrixMultiply(&pFrame, &pFrame, &getProjectionFrame());
        return pFrame;
    }

    /**  Edit the projection frame.
     */
    void setProjectionFrame(float pAperture, float pAspectRatio, float pNear, float pFar)
    {
        mCamProjection.mAperture = pAperture;
        mCamProjection.mAspectRatio = pAspectRatio;
        mCamProjection.mNear = pNear;
        mCamProjection.mFar = pFar;
        updateProjectionFrame();
    }
    void setAperture(float pAperture)
    {
        mCamProjection.mAperture = pAperture;
        updateProjectionFrame();
    }
    void setAspectRatio(float pAspectRatio)
    {
        mCamProjection.mAspectRatio = pAspectRatio;
        updateProjectionFrame();
    }
    void setNear(float pNear)
    {
        mCamProjection.mNear = pNear;
        updateProjectionFrame();
    }
    void setFar(float pFar)
    {
        mCamProjection.mFar = pFar;
        updateProjectionFrame();
    }
    float getAperture() const {return mCamProjection.mAperture;}
    float getAspectRatio() const {return mCamProjection.mAspectRatio;}
    float getNear() const {return mCamProjection.mNear;}
    float getFar() const {return mCamProjection.mFar;}
    float getHeight() const {return mCamProjection.mNear*2.f*tan(0.5f*mCamProjection.mAperture);}
    float getWidth() const {return getHeight() * getAspectRatio();}

    /**  Edit the view frame.
     */
    void setViewFrame(D3DXVECTOR3& pInspectPoint, D3DXVECTOR3& pEuler, float pDistance, bool pResetPos = true)
    {	
        //Reset all view frames attributes
        D3DXMatrixIdentity(&mInspectFrame);	
        mCamLocalPos.z = -abs(pDistance);
        mCamLocalPos.x=0;
        mCamLocalPos.y=0;
        mCamRotate = pEuler;
        mCamTranslate = D3DXVECTOR3(0, 0, 0);
        mCamInspectPoint = pInspectPoint;

        if (pResetPos)
        {
            mCamInitInspectPoint = pInspectPoint;
            mCamInitEuler = pEuler;
            mCamInitZoom = pDistance;
        }

        // Then update.
        updateViewFrame();
    }

    /**
     * Reset the view frame to the last assign pos through setViewFrame().
     */
    void resetViewFrame()
    {
        setViewFrame(mCamInitInspectPoint, mCamInitEuler, mCamInitZoom, false);
    }

    /**  Edit the model frame.
     */
    D3DXMATRIX& editWorldFrame() {return mWorldFrame;}

    /**  Event handler method.
     */
    virtual bool handleMessages( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
    {
        // Evaluate Window Message
        switch (uMsg)														
        {
            case WM_MOUSEWHEEL:
            {
                CkState eventState( GET_BUTTONSTATE( wParam ),
                                    GET_MOUSECURSORX( lParam ), 
                                    GET_MOUSECURSORY( lParam ),
                                    GET_WHEEL_DELTA_WPARAM( wParam ) );
                mCamLocalPos.z += (eventState.mMouseWheelDelta) * mCamSpeed.z * abs(mCamLocalPos.z) / 120.f;
                if (mCamLocalPos.z > -0.01f)
                {
                    mCamLocalPos.z = -0.01f;
                }
                updateViewFrame();
                return true;
            }
            break;
            case WM_MOUSEMOVE:
            {
                CkState eventState( GET_BUTTONSTATE( wParam ),
                                    GET_MOUSECURSORX( lParam ), 
                                    GET_MOUSECURSORY( lParam ) );
                bool lReturn = false;
                
                if (eventState.mButtonState == MK_MBUTTON )
                {
                    mCamTranslate.z -= (eventState.mMouseCursorY - mInputState.mMouseCursorY) * mCamSpeed.x * abs(mCamLocalPos.z);
                    
                    lReturn =  true;
                }
                else if	(eventState.mButtonState & MK_RBUTTON)
                {
                    mCamTranslate.x += (eventState.mMouseCursorX - mInputState.mMouseCursorX) * mCamSpeed.x * abs(mCamLocalPos.z);
                    mCamTranslate.y -= (eventState.mMouseCursorY - mInputState.mMouseCursorY) * mCamSpeed.x * abs(mCamLocalPos.z);
                    lReturn =  true;
                }
                else if	(eventState.mButtonState & MK_LBUTTON)
                {
                    mCamRotate.x -= (eventState.mMouseCursorX - mInputState.mMouseCursorX) * mCamSpeed.y;
                    mCamRotate.y -= (eventState.mMouseCursorY - mInputState.mMouseCursorY) * mCamSpeed.y;
                    mCamRotate.y = min(mCamRotate.y, (1.56f));
                    mCamRotate.y = max(mCamRotate.y, -(1.56f));
                    lReturn =  true;
                }

                if (lReturn == true)
                {
                    updateViewFrame();

                    mInputState.mMouseCursorX = eventState.mMouseCursorX;
                    mInputState.mMouseCursorY = eventState.mMouseCursorY;	
                }
                return lReturn;
            }
            break; 
            case WM_LBUTTONDOWN:
            case WM_RBUTTONDOWN:
            case WM_MBUTTONDOWN:
            {
                CkState eventState( GET_BUTTONSTATE( wParam ),
                                    GET_MOUSECURSORX( lParam ), 
                                    GET_MOUSECURSORY( lParam ) );
                mInputState.mButtonState = eventState.mButtonState;
                mInputState.mMouseCursorX = eventState.mMouseCursorX;
                mInputState.mMouseCursorY = eventState.mMouseCursorY;
                return false;
            }
            break;
            case WM_LBUTTONUP:
            case WM_RBUTTONUP:
            case WM_MBUTTONUP:
            {
                CkState eventState( GET_BUTTONSTATE( wParam ),
                                    GET_MOUSECURSORX( lParam ), 
                                    GET_MOUSECURSORY( lParam ) );
                mInputState.mButtonState = eventState.mButtonState;
                return false;
            }
            break;
            case WM_LBUTTONDBLCLK:
            case WM_MBUTTONDBLCLK:
            case WM_RBUTTONDBLCLK:
            {
                return false;
            }
            break;
        }

        return false;
    }

protected:
    /******************************* ATTRIBUTES *******************************/

    /** The projection frame state.
     */
    CkProjection mCamProjection;
    D3DXMATRIXA16 mProjectionFrame;

    /** The view frame state.
     */
    D3DXMATRIXA16 mViewFrame;
    D3DXMATRIXA16 mInspectFrame;


    /** The model frame state.
     */
    D3DXMATRIXA16 mWorldFrame;

    /** The current input state of the Camera.
     * It is updated by the different events occuring.
     */	
    CkState mInputState;

    /** The current mode of the camera manipulator.
     */
    CkMode mMode;

    /**
     * The camera parameters and the starting pos are stored so the view can be reset.
     */
    D3DXVECTOR3 mCamInspectPoint;
    D3DXVECTOR3 mCamLocalPos;
    D3DXVECTOR3 mCamTranslate;
    D3DXVECTOR3 mCamRotate;	
    D3DXVECTOR3 mCamSpeed;
    D3DXVECTOR3 mCamInitInspectPoint;
    D3DXVECTOR3 mCamInitEuler;
    float mCamInitZoom;


    /******************************** METHODS *********************************/

    /** Update the projection frame from the projection attributes.
     */
    void updateProjectionFrame()
    {
        D3DXMatrixPerspectiveFovLH( &mProjectionFrame,
                                    mCamProjection.mAperture, 
                                    mCamProjection.mAspectRatio, 
                                    mCamProjection.mNear, 
                                    mCamProjection.mFar );
    }

    /** Update the view frame from the camera attributes.
     */
    void updateViewFrame()
    {
        switch (mMode)
        {
        case eInspect:
            {
                // Temp matrix used first for rotation, then for translation.
                D3DXMATRIXA16 lMat;
                D3DXMATRIXA16 lMatInspectTI;
                
                // rotate arc ball
                D3DXMatrixRotationYawPitchRoll( &lMat, mCamRotate.x , mCamRotate.y , mCamRotate.z );	
                D3DXMatrixMultiply(&mInspectFrame, &mInspectFrame, &lMat);
                D3DXMatrixInverse(&lMatInspectTI, NULL, &mInspectFrame);

                // compute the translation from view space to world space
                D3DXVECTOR4  lCamMove;
                D3DXVec3Transform(&lCamMove, &mCamTranslate, &lMatInspectTI);
                mCamInspectPoint.x += lCamMove.x;
                mCamInspectPoint.y += lCamMove.y;
                mCamInspectPoint.z += lCamMove.z;

                // Then recompute the view frame = CamLookAtPoint + InspectFrame + mCamLocalPos.z
                D3DXMatrixTranslation(&mViewFrame, mCamInspectPoint.x, mCamInspectPoint.y, mCamInspectPoint.z);
                D3DXMatrixMultiply(&mViewFrame, &mViewFrame, &mInspectFrame);
                D3DXMatrixTranslation(&lMat, 0, 0, -mCamLocalPos.z);
                D3DXMatrixMultiply(&mViewFrame, &mViewFrame, &lMat);

                // reset commands.
                mCamTranslate.x = mCamTranslate.y = mCamTranslate.z = 0;
                mCamRotate.x = mCamRotate.y = mCamRotate.z = 0;
            }
            break;
        case eTerrain:
            {

                // Compute the LookAt position by moving it with the LookAtMove transformed in world.
                D3DXMatrixRotationYawPitchRoll( &mViewFrame, -mCamRotate.x , 0 , 0 );	
                D3DXVECTOR4  lLookAtMove(-mCamTranslate.x, -mCamTranslate.z, (mCamRotate.y>=0?1:-1)*mCamTranslate.y, 0);
                D3DXVec4Transform(&lLookAtMove, &lLookAtMove, &mViewFrame);	
                mCamInspectPoint.x += lLookAtMove.x;
                mCamInspectPoint.y += lLookAtMove.y;
                mCamInspectPoint.z += lLookAtMove.z;
                mCamTranslate = D3DXVECTOR3(0, 0, 0);

                // Compute the CamPos in world from the euler angles and the Cam local position.
                D3DXMatrixRotationYawPitchRoll( &mViewFrame, -mCamRotate.x , -mCamRotate.y , mCamRotate.z );
                D3DXVECTOR4  lCamPos;
                D3DXVec3Transform(&lCamPos, &mCamLocalPos, &mViewFrame);
                D3DXVECTOR3 lEyePt( lCamPos.x, lCamPos.y, lCamPos.z);
                lEyePt.x += mCamInspectPoint.x;
                lEyePt.y += mCamInspectPoint.y;
                lEyePt.z += mCamInspectPoint.z;

                // Compute the final cam frame from the state vec.
                D3DXVECTOR3 lUpVec( 0.0f, 1.0f, 0.0f );
                D3DXMatrixLookAtLH( &mViewFrame, &lEyePt, &mCamInspectPoint, &lUpVec);
            }
            break;
        default:
            break;
        }
    }
    
    /** Update the world frame, simply force identity.
     */
    void updateWorldFrame()
    {
        D3DXMatrixIdentity(&mWorldFrame);
    }
};



/** CkDemoWindow
 * Class managing the mozaic viewports and a selected vieport full pWindow.
 * for the different techniques displayed.
 *
 * Also display text integrating the CDXUT
 */
class CkDemoWindow
{
public:
    /********************************* TYPES **********************************/
    
    typedef void *CkAnimCallback(double);
    typedef void *CkResetCallback();
    typedef void *CkLoadModelCallback();
    
    /**
     * CkProperty inner class make the bridge between a mVariable (pure type or from an FX) and
     * its GUI displayed in the pWindow.
     * CkProperty is the base pure virtual class, create only specialisations.
     */
    class CkProperty
    {
    public:
        UINT mWidgetID;
        UINT mLabelID;
        CDXUTStatic* mLabel;

        virtual void updateVar() = 0;
		virtual void reset() = 0;

    protected:
        CkProperty() :
             mWidgetID(0),
             mLabelID(0),
             mLabel(NULL)
        {}
    };

    class CkPropertyFXScalar : public CkProperty
    {
    public:
        CDXUTSlider* mWidget;
        ID3D10EffectScalarVariable* mVariable;
		float mResetVal;
        std::wstring mName;
        float mScale;
        float mOffset;
        
        CkPropertyFXScalar(CkDemoWindow& pWindow, ID3D10EffectScalarVariable* pVar, float pMin, float pMax, std::wstring& pName) :
            CkProperty(),
            mVariable(pVar),
            mName(pName)
        {
            int yPos = pWindow.addProperty(this);
            
            mScale = (pMax - pMin) / 100;
            mOffset = pMin;

            float value;
            mVariable->GetFloat(&value);
            mResetVal = value;

            WCHAR sz[100];
            StringCchPrintf( sz, 100, L"%s : %0.3f", mName.c_str(), value ); 
            pWindow.mDXPropertyUI.AddStatic( mLabelID, sz, (int) (0.4f * CkDemoWindow::sControlWidth+2.f), yPos, (int) (0.6f * CkDemoWindow::sControlWidth - 2.f), CkDemoWindow::sControlHeight, false, &mLabel);

            pWindow.mDXPropertyUI.AddSlider(mWidgetID, 0, yPos, (int) (0.4f * CkDemoWindow::sControlWidth), CkDemoWindow::sControlHeight, 0, 100,  static_cast<int> ((value - mOffset)/mScale), false, &mWidget );
            mWidget->SetUserData(this);

        }

        virtual void updateVar()
        {
            float value = mWidget->GetValue() * mScale + mOffset;
            mVariable->SetFloat(value);
            WCHAR sz[100];
            StringCchPrintf( sz, 100, L"%s : %0.3f", mName.c_str(), value ); 
            mLabel->SetText(sz );
        }

        virtual void reset()
        {
			mWidget->SetValue(static_cast<int> ((mResetVal - mOffset)/mScale) );
            WCHAR sz[100];
            StringCchPrintf( sz, 100, L"%s : %0.3f", mName.c_str(), mResetVal ); 
            mLabel->SetText(sz );			
            mVariable->SetFloat(mResetVal);
        }

    };

    class CkPropertyBool : public CkProperty
    {
    public:
        CDXUTCheckBox* mWidget;
        bool* mVariable;
        bool  mResetVal;
        std::wstring mName;
        
        CkPropertyBool(CkDemoWindow& pWindow, bool* pVar, std::wstring& pName) :
            CkProperty(),
            mVariable(pVar),
            mResetVal(*pVar),
            mName(pName)
        {
            int yPos = pWindow.addProperty(this);
            pWindow.mDXPropertyUI.AddCheckBox(mWidgetID, mName.c_str(), 0, yPos, CkDemoWindow::sControlWidth, CkDemoWindow::sControlHeight, (*mVariable), 0, false, &mWidget );
            mWidget->SetUserData(this);
        }

        virtual void updateVar()
        {
            (*mVariable) = mWidget->GetChecked();
        }

        virtual void reset()
        {
			mWidget->SetChecked(mResetVal);
            (*mVariable) = mResetVal;
        }

    };

    class CkPropertyEnum : public CkProperty
    {
    public:
        CDXUTComboBox* mWidget;
        int* mVariable;
		int  mResetVal;
        std::wstring mName;
        
        CkPropertyEnum(CkDemoWindow& pWindow, int* pVar, std::vector< std::wstring >& pItems, std::wstring& pName) :
            CkProperty(),
            mVariable(pVar),
			mResetVal(*pVar),
            mName(pName)
        {
            int yPos = pWindow.addProperty(this);
                        
            pWindow.mDXPropertyUI.AddComboBox(mWidgetID, 0, yPos, CkDemoWindow::sControlWidth, CkDemoWindow::sControlHeight, 0, false, &mWidget );
            mWidget->SetUserData(this);
            for (std::vector< std::wstring >::iterator item = pItems.begin(); item != pItems.end(); ++item)
            {
                mWidget->AddItem((*item).c_str(), NULL);
            }
            mWidget->SetSelectedByIndex(*mVariable);
        }

        virtual void updateVar()
        {
            (*mVariable) = mWidget->GetSelectedIndex();
        }

        virtual void reset()
        {
            mWidget->SetSelectedByIndex(mResetVal);
            (*mVariable) = mResetVal;
        }

	};

    /******************************* ATTRIBUTES *******************************/
    
    CkCamera    mCamera;
    int     mNbViewportsX;
    int     mNbViewportsY;
    int     mWindowWidth;
    int     mWindowHeight;
    int     mSelectedViewport;
    int     mPointedViewport;
    std::vector< std::wstring > mViewportTitles;

    ID3D10RasterizerState*      mDX10RasterizerState;
    ID3DX10Font*                mDX10Font;
    CDXUTDialogResourceManager  mDXDialogResourceManager;
    CD3DSettingsDlg             mD3DSettingsDlg;
    CDXUTDialog                 mDXHUD;
    CDXUTDialog                 mDXPropertyUI;
    std::vector< CkProperty* >  mProperties;
    UINT mWidgetID;
    bool mGUICaptureMouse;
    CkResetCallback*			mResetCallback;
    CkLoadModelCallback*		mLoadModelCallback;

    bool mDisplayHelp;
    bool mDisplayInfo;
    bool mDisplayStats;
    bool mDisplayGUI;
    bool mCaptureSnap;
    bool mCaptureAnim;
    int  mCaptureNb;
    int  mCaptureAnimNb;
    CkAnimCallback* mAnimCallback;


    static const int sTextHeight = 15;
    static const int sControlWidth = 200;
    static const int sControlHeight = 20;
    static const int sControlSpace = 2;

    //--------------------------------------------------------------------------------------
    // UI control IDs
    //--------------------------------------------------------------------------------------
    enum UIcontrolIDs
    {   
        IDC_TOGGLEFULLSCREEN = 1,
        IDC_TOGGLEREF,
        IDC_CHANGEDEVICE,
        IDC_RESET,
        IDC_LOADMODEL,
    };

    /******************************** METHODS *********************************/

    /** Constructor - Destructor.
     */
    CkDemoWindow() :
        mCamera(CkCamera::eInspect),
        mNbViewportsX(2),
        mNbViewportsY(3),
        mWindowWidth(1),
        mWindowHeight(1),
        mSelectedViewport(-1),
        mPointedViewport(-1),
        mDX10RasterizerState(NULL),
        mDX10Font(NULL),
        mWidgetID(0),
        mGUICaptureMouse(false),
        mDisplayHelp(false),
        mDisplayInfo(true),
        mDisplayStats(true),
        mDisplayGUI(true),
        mCaptureSnap(false),
        mCaptureAnim(false),
        mCaptureNb(0),
        mCaptureAnimNb(0),
        mAnimCallback(NULL),
        mResetCallback(NULL),
        mLoadModelCallback(NULL)
    {
        mDXPropertyUI.Init( &mDXDialogResourceManager );
        mDXPropertyUI.SetCallback(CkDemoWindow::OnGUIEvent);

        mD3DSettingsDlg.Init( &mDXDialogResourceManager );

        mDXHUD.Init( &mDXDialogResourceManager );
        mDXHUD.SetCallback( CkDemoWindow::OnGUIEventID ); 
        int iY = 10; 
        mDXHUD.AddButton( IDC_TOGGLEFULLSCREEN, L"Toggle full screen (F2)", 35, iY, 125, sControlHeight, VK_F2);

        CDXUTButton* lButton = NULL;
        mDXHUD.AddButton( IDC_CHANGEDEVICE, L"Change device (F3)", 35, iY += sControlHeight+sControlSpace, 125, sControlHeight, VK_F3, false, &lButton);
        if (lButton)
        {
            lButton->SetUserData(this);
        }
 //       mDXHUD.AddButton( IDC_TOGGLEREF, L"Toggle REF (F3)", 35, iY += sControlHeight+sControlSpace, 125, sControlHeight, VK_F3 );
        mDXHUD.AddButton( IDC_RESET, L"Reset (F4)", 35, iY += sControlHeight+sControlSpace, 125, sControlHeight, VK_F4, false, &lButton);
        if (lButton)
        {
            lButton->SetUserData(this);
        }
        mDXHUD.AddButton( IDC_LOADMODEL, L"Load model (F5)", 35, iY += sControlHeight+sControlSpace, 125, sControlHeight, VK_F5, false, &lButton);
        if (lButton)
        {
            lButton->SetUserData(this);
        }
    }

    ~CkDemoWindow()
    {
    }

    float getMozaicViewportWidth()
    {
        return (static_cast<float> (mWindowWidth) / static_cast<float> (mNbViewportsX));
    }
    float getMozaicViewportHeight()
    {
        return (static_cast<float> (mWindowHeight) / static_cast<float> (mNbViewportsY));
    }

    float getViewportWidth(int v)
    {
        if (mSelectedViewport < 0)
        {
            return getMozaicViewportWidth();
        }
        else if (mSelectedViewport == v)
        {
            return static_cast<float> (mWindowWidth);
        }
        else
        {
            return 0.f;
        }
    }
    float getViewportHeight(int v)
    {
        if (mSelectedViewport < 0)
        {
            return getMozaicViewportHeight();
        }
        else if (mSelectedViewport == v)
        {
            return static_cast<float> (mWindowHeight);
        }
        else
        {
            return 0;
        }
    }	
    float getViewportAspectRatio(int v)
    {
        if (mSelectedViewport < 0)
        {
            return getMozaicViewportWidth()/getMozaicViewportHeight();
        }
        else 
        {
            return (static_cast<float> (mWindowWidth) / static_cast<float> (mWindowHeight));
        }
    }
    float getViewportTopLeftX(int v)
    {
        if (mSelectedViewport < 0)
        {
            return (v % mNbViewportsX) * getMozaicViewportWidth();
        }
        else 
        {
            return 0;
        }
    }
    float getViewportTopLeftY(int v)
    {
        if (mSelectedViewport < 0)
        {
            return v/mNbViewportsX * getMozaicViewportHeight();
        }
        else 
        {
            return 0;
        }
    }

    int getViewportIndex(int x, int y)
    {
        if (x < 0 || x > mWindowWidth) return -1;
        if (y < 0 || y > mWindowHeight) return -1;
        if (mSelectedViewport < 0)
        {
            int rawId = x / static_cast<int>(getMozaicViewportWidth());
            int colId = y / static_cast<int>(getMozaicViewportHeight());
            return colId*mNbViewportsX + rawId;
        }
        else
        {
            return mSelectedViewport;
        }
    }

    void setNbViewports(int nbVs)
    {
        float root = sqrt( static_cast<float> (nbVs) );
        mNbViewportsX = static_cast<int> (ceil(root));
        mNbViewportsY = static_cast<int> (ceil(static_cast<float> (nbVs) / mNbViewportsX));
    }
    
    void selectViewport(int selected)
    {
        mSelectedViewport = selected;
    }

    void setResetCallback(CkResetCallback* pResetCallback)
    {
        mResetCallback = pResetCallback;
    }

	void setLoadModelCallback(CkLoadModelCallback* pLoadModelCallback)
    {
        mLoadModelCallback = pLoadModelCallback;
    }

    /******************************** Propagate Callbacks *********************************/

    HRESULT OnD3D10CreateDevice( ID3D10Device* pd3dDevice, const DXGI_SURFACE_DESC* pBackBufferSurfaceDesc)
    {
        HRESULT hr;
        
        V_RETURN( D3DX10CreateFont( pd3dDevice, sTextHeight, 0, FW_LIGHT, 1, FALSE, DEFAULT_CHARSET, 
                                    OUT_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH | FF_DONTCARE, 
                                    L"Arial", &mDX10Font ) );

        V_RETURN( mDXDialogResourceManager.OnD3D10CreateDevice( pd3dDevice ) );
        V_RETURN( mD3DSettingsDlg.OnD3D10CreateDevice( pd3dDevice ) );


        D3D10_RASTERIZER_DESC rasterizerState;
        rasterizerState.FillMode = D3D10_FILL_SOLID;
        rasterizerState.CullMode = D3D10_CULL_NONE;
        rasterizerState.FrontCounterClockwise = true;
        rasterizerState.DepthBias = false;
        rasterizerState.DepthBiasClamp = 0;
        rasterizerState.SlopeScaledDepthBias = 0;
        rasterizerState.DepthClipEnable = false;
        rasterizerState.ScissorEnable = false;
        rasterizerState.MultisampleEnable = true;
        rasterizerState.AntialiasedLineEnable = false;
        pd3dDevice->CreateRasterizerState( &rasterizerState, &mDX10RasterizerState );

        return S_OK;
    }
                
    void OnD3D10ResizedSwapChain( ID3D10Device* pd3dDevice, IDXGISwapChain *pSwapChain, const DXGI_SURFACE_DESC* pBackBufferSurfaceDesc )
    {
        mWindowWidth = pBackBufferSurfaceDesc->Width;
        mWindowHeight = pBackBufferSurfaceDesc->Height;
         
        mDXDialogResourceManager.OnD3D10ResizedSwapChain( pd3dDevice, pBackBufferSurfaceDesc );
        mD3DSettingsDlg.OnD3D10ResizedSwapChain( pd3dDevice, pBackBufferSurfaceDesc );

        mDXHUD.SetLocation( pBackBufferSurfaceDesc->Width-170, 0 );
        mDXHUD.SetSize( 170, 170 );

        mDXPropertyUI.SetLocation(mWindowWidth - sControlWidth - 5, computePropertyUILocationY());
        mDXPropertyUI.SetSize( sControlWidth, computePropertyUIHeight());
    }
    void OnD3D10ReleasingSwapChain()
    {
        mDXDialogResourceManager.OnD3D10ReleasingSwapChain();
    }

    void OnD3D10DestroyDevice()
    {
        clearProperties();

        mDXDialogResourceManager.OnD3D10DestroyDevice();
        mD3DSettingsDlg.OnD3D10DestroyDevice();

        SAFE_RELEASE( mDX10RasterizerState );
        SAFE_RELEASE( mDX10Font );
    }

    void resetProperties()
    {
        for (unsigned int editorId = 0; editorId < mProperties.size(); ++editorId)
        {
            mProperties[editorId]->reset();
        }
    }
	
	void clearProperties()
    {
        mDXPropertyUI.RemoveAllControls();
        for (unsigned int editorId = 0; editorId < mProperties.size(); ++editorId)
        {
            delete mProperties[editorId];
        }
        mProperties.clear();
    }

    void OnD3D10FrameRender(ID3D10Device* pd3dDevice, double fTime, float fElapsedTime)
    {
        // If the settings dialog is being shown, then
        // render it instead of rendering the app's scene
        if( mD3DSettingsDlg.IsActive() )
        {
            // Update the viewport to draw in the full pWindow viewport. 
            D3D10_VIEWPORT vp;
            vp.Width = mWindowWidth;
            vp.Height = mWindowHeight;
            vp.MinDepth = 0;
            vp.MaxDepth = 1;
            vp.TopLeftX =  0;
            vp.TopLeftY = 0;
            pd3dDevice->RSSetViewports( 1, &vp );

            // Force the Rasterizer state.
            pd3dDevice->RSSetState(mDX10RasterizerState);

            mD3DSettingsDlg.OnRender( fElapsedTime );
            return;
        }

        //Before any UI, capture snap or video animation
        if ( mCaptureAnim)
        {
            captureFrame();

            if (mAnimCallback)
            {
                mAnimCallback(fTime);
            }
            mCaptureSnap = false;
            ++mCaptureAnimNb;
            if (mCaptureAnimNb > 30)
            {
                mCaptureAnim = false;
                mCaptureAnimNb = 0;
            }
        }

        if (mCaptureSnap)
        {
            captureFrame();
            mCaptureSnap = false;
        }

        // Update the viewport to draw in the full pWindow viewport. 
        D3D10_VIEWPORT vp;
        vp.Width = mWindowWidth;
        vp.Height = mWindowHeight;
        vp.MinDepth = 0;
        vp.MaxDepth = 1;
        vp.TopLeftX =  0;
        vp.TopLeftY = 0;
        pd3dDevice->RSSetViewports( 1, &vp );

        // Force the Rasterizer state.
        pd3dDevice->RSSetState(mDX10RasterizerState);

        renderUI(pd3dDevice, fTime, fElapsedTime);
        renderText(pd3dDevice, fTime, fElapsedTime);
    }

    void renderUI(ID3D10Device* pd3dDevice, double fTime, float fElapsedTime)
    {
        mDXHUD.OnRender( fElapsedTime );
        if (mDisplayGUI) mDXPropertyUI.OnRender( fElapsedTime ); 
    }

    void renderText(ID3D10Device* pd3dDevice, double fTime, float fElapsedTime)
    {
        CDXUTTextHelper txtHelper(mDX10Font, NULL, sTextHeight );

        if (mDisplayStats)
        {
            txtHelper.Begin();
            txtHelper.SetInsertionPos( 5, 5 );
            txtHelper.SetForegroundColor( D3DXCOLOR( 1.0f, 1.0f, 0.0f, 1.0f ) );
            txtHelper.DrawTextLine( DXUTGetFrameStats( true ) ); 
            txtHelper.DrawTextLine( DXUTGetDeviceStats() );
            txtHelper.End();
       }

        // Draw help or "F1 forHelp"
        txtHelper.Begin();
        txtHelper.SetForegroundColor( D3DXCOLOR( 1.0f, 1.0f, 1.0f, 1.0f ) );
        txtHelper.SetInsertionPos( 5, 5+2*sTextHeight);
        if (mDisplayHelp)
        {
            txtHelper.DrawTextLine(
                L"- ESC to quit\n"
                L"- F1 to show/hide help\n"
                L"- F4 to reset the camera to the initial position\n"
                L"- F5 to load a model\n"
                L"- F6 to show/hide statistics and device info\n"
                L"- F7 to show/hide property editors\n"
                L"- F8 to show/hide selected technique info\n"
                L"- F9 to capture a snapshot\n"
                );
            txtHelper.SetInsertionPos( 5, 5+9*sTextHeight);
            txtHelper.DrawTextLine(
                L"\n"
                L"- Left drag to orbit around pivot point\n"
                L"- Right drag to pan pivot point\n"
                L"- Middle drag to translate pivot point in depth\n"
                L"- Wheel to zoom in/out camera to pivot point\n"
                L"\n"
                L"- Left Double click on a viewport to swith\n"
                L"  between full window and mozaic view\n"
                );
        }
        else
        {
            txtHelper.DrawTextLine( L"Press F1 for help" );
        }
        txtHelper.End();
        
        if (mDisplayInfo)
        {
            int lDisplayedTitle = mPointedViewport;
            if  (lDisplayedTitle >=0 && (int) mViewportTitles.size() > lDisplayedTitle)
            {
                txtHelper.Begin();
                txtHelper.SetInsertionPos( 5, mWindowHeight - 5 -sTextHeight );
                txtHelper.SetForegroundColor( D3DXCOLOR( 0.4f, 1.f, 0.1f, 1.0f ) );
                txtHelper.DrawTextLine(mViewportTitles[lDisplayedTitle].c_str());
                txtHelper.End();
            }
        }

    }

    void captureFrame()
    {
        IDXGISwapChain* pSwapChain = DXUTGetDXGISwapChain();
        ID3D10Texture2D* pRT;
        pSwapChain->GetBuffer(0, __uuidof(pRT), reinterpret_cast<void**>(&pRT));
        WCHAR filename[32];
        StringCchPrintf(filename, 100, L"screenshot%.3d.dds", mCaptureNb); 
        D3DX10SaveTextureToFile(pRT, D3DX10_IFF_DDS, filename);
        pRT->Release();
        ++mCaptureNb;
    }

    void onReset()
    {
        mCamera.resetViewFrame();
		resetProperties();
		if (mResetCallback) mResetCallback();
    }

	void onLoadModel()
    {
        if (mLoadModelCallback) mLoadModelCallback();
    }

    bool MsgProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
    {
        g_hWnd = hWnd;
        // Pass messages to dialog resource manager calls so GUI state is updated correctly
        if( mDXDialogResourceManager.MsgProc( hWnd, uMsg, wParam, lParam ) )
            return true;  
        // Pass messages to settings dialog if its active
        if( mD3DSettingsDlg.IsActive() )
        {
            mD3DSettingsDlg.MsgProc( hWnd, uMsg, wParam, lParam );
            return 0;
        }

        // Give the dialogs a chance to handle the message first
        if( mDXHUD.MsgProc( hWnd, uMsg, wParam, lParam ) )
        {
            mGUICaptureMouse = true;
            return true;
        }

        if( mDisplayGUI && mDXPropertyUI.MsgProc( hWnd, uMsg, wParam, lParam ) )
        {
            mGUICaptureMouse = true;
            return true;
        }
        
        // Update the cviewport currently pointed by the mouse, -1 when loosing focus.
        if(uMsg == WM_MOUSEMOVE)
        {
            int lMouseX = (short)LOWORD(lParam);
            int lMouseY = (short)HIWORD(lParam);
            mPointedViewport = getViewportIndex(lMouseX, lMouseY);
        }

        // When double click, toggle FullWindow / Mosaic mode
        if(uMsg == WM_LBUTTONDBLCLK)
        {
            int lMouseX = (short)LOWORD(lParam);
            int lMouseY = (short)HIWORD(lParam);
            if (mSelectedViewport < 0)
            {
                selectViewport( getViewportIndex(lMouseX, lMouseY) );
            }
            else
            {
                selectViewport( -1 );
            }
            return true;
        }

        // Consolidate the keyboard messages and pass them to the app's keyboard callback
        if( uMsg == WM_KEYDOWN ||
            uMsg == WM_SYSKEYDOWN || 
            uMsg == WM_KEYUP ||
            uMsg == WM_SYSKEYUP )
        {
            bool quit = false;

            bool bKeyDown = (uMsg == WM_KEYDOWN || uMsg == WM_SYSKEYDOWN);

            UINT nChar = (UINT)wParam;

            if( bKeyDown )
            {
                switch( nChar )
                {
                    case VK_F1:
                    {
                        mDisplayHelp = !mDisplayHelp;
                        quit = true;
                    }
                    break;
                    case VK_F6:
                    {
                        mDisplayStats = !mDisplayStats;
                        quit = true;
                    }
                    break;
                    case VK_F7:
                    {
                        mDisplayGUI = !mDisplayGUI;
                        quit = true;
                    }
                    break;
                    case VK_F8:
                    {
                        mDisplayInfo = !mDisplayInfo;
                        quit = true;
                    }
                    break;
                    case VK_F9:
                    {
                        mCaptureSnap = true;
                        quit = true;
                    }
                    /*case VK_F10:
                    {
                        mCaptureAnim = true;
                        quit = true;
                    }*/
                    break;
                }

                if (quit) return true;
            }
        }

        // Pass all remaining windows messages to camera so it can respond to user input except if GUI Captured Mouse.
        if (!mGUICaptureMouse)
        {
            return mCamera.handleMessages( hWnd, uMsg, wParam, lParam );
        }
        // GUI free MouseEvent on button up.
        else if(    uMsg == WM_LBUTTONUP ||
                    uMsg == WM_RBUTTONUP || 
                    uMsg == WM_MBUTTONUP ) 
        {
            mGUICaptureMouse = false;
        }
        
        return false;
    }

    static void CALLBACK OnGUIEvent( UINT nEvent, int nControlID, CDXUTControl* pControl, void* pUserContext )
    {
        if ( pControl->GetUserData())
        {
            CkProperty* editor = static_cast<CkProperty*> (pControl->GetUserData());
            editor->updateVar();
        }

    }

    static void CALLBACK OnGUIEventID( UINT nEvent, int nControlID, CDXUTControl* pControl, void* pUserContext )
    {
        switch( nControlID )
        {
            case IDC_TOGGLEFULLSCREEN:
                PostMessageW(DXUTGetHWND(), WM_SYSKEYDOWN, VK_RETURN, (1 << 29));
                 //DXUTToggleFullScreen();
                break;
            case IDC_TOGGLEREF:
                DXUTToggleREF();
                break;
            case IDC_CHANGEDEVICE:
                {
                    CkDemoWindow* lWindow = static_cast<CkDemoWindow*>(pControl->GetUserData());
                    lWindow->mD3DSettingsDlg.SetActive( !lWindow->mD3DSettingsDlg.IsActive() ); 
                }
                break;
            case IDC_RESET:
                {
                    CkDemoWindow* lWindow = static_cast<CkDemoWindow*>(pControl->GetUserData());
					lWindow->onReset();
                }
                break;
            case IDC_LOADMODEL:
                {
                    CkDemoWindow* lWindow = static_cast<CkDemoWindow*>(pControl->GetUserData());
                    lWindow->onLoadModel();
                }
                break;
        }
    }

    int addProperty(CkProperty* prop)
    {
        int UIY = computePropertyUIHeight();

        //back up the mWidget ID.
        mWidgetID++;
        prop->mLabelID = mWidgetID;
        mWidgetID++;
        prop->mWidgetID = mWidgetID;
    
        mProperties.push_back(prop);

        // Move the propertyUI display one up
        mDXPropertyUI.SetLocation(mWindowWidth - sControlWidth - 5, computePropertyUILocationY());

        return UIY;
    }

    int computePropertyUIHeight()
    {
        return (mWidgetID / 2) * (sControlHeight + sControlSpace);
    }
    int computePropertyUILocationY()
    {
        return  mWindowHeight - computePropertyUIHeight();
    }
};